#!/usr/bin/perl -w

# BruteForceBlocker.pl v1.1
#
# This is a new version on BruteForceBlocker, that is a way
# faster, less demanding (we don't need File::Tail anymore)
# and catches more entries. Enjoy!
#
# Thanks to Jonas Davidsson and Matt Pearce for good ideas.
#
#
# HOWTO:
#
# You will need to add a new table to the pf config file
# you can do so by adding lines to the pf.conf similar to these:
#
# table <bruteforce> persist file "/path/to/table"
# block in log quick inet proto tcp from <bruteforce> to any port ssh
#
# You will also need to modify auth line in /etc/syslog.conf so
# it will look like this:
#
# auth.info;authpriv.info	| exec /path/to/bruteforceblocker.pl
#
# You should consider starting syslogd with -c option as well.
#
# - Daniel Gerzo <danger@rulez.sk>

use strict;
use warnings;

my $logfile		= '/var/log/auth.log';	# file where this script will log to
my $pfctl		= '/sbin/pfctl';	# pfctl binary
my $table		= 'bruteforce';		# pf table
my $tablefile		= '/path/to/table';	# file where table persist
my $max_attempts	= 3;			# number of max allowed fails
my $timeout		= 3600;			# number of seconds after resetting of ip
my %count		= ();			# number of failed tries
my %time		= ();			# last modified time

# list of IPs that will never be blocked
my @trusted_IPs = qw(
	10.10.1.72
	10.1.255.254
);

# open log for writing and disable caching
open (FILE, ">>$logfile") or die ("Couldn't open $logfile for writing");
select FILE; $| = 1; select STDOUT;

print (FILE "\n------- log started at ", scalar(localtime), " -------\n\n");

# the core process
while (<>) {
	print (FILE "$_");

	if (/.*Failed password.*from ((?:\d{1,3}\.){3}\d{1,3}) port.*/ or
		/.*Invalid user.*from ((?:\d{1,3}\.){3}\d{1,3}).*/ or
		/.*User.*from ((?:\d{1,3}\.){3}\d{1,3}) not allowed because not listed in AllowUsers/) {

		my $IP = $1;

		$count{$IP}++;

		print (FILE "$IP was logged with total count of $count{$IP}.\n") if ($count{$IP} < $max_attempts+1);
		
		if($time{$IP}) {
			if ($time{$IP} < time - $timeout) {
				print (FILE "resetting $IP count, since it wasn't modified for more than $timeout seconds.\n");
				delete $count{$IP};
			}
		}
		
		$time{$IP} = time;

		if($count{$IP} == $max_attempts+1) {
			print (FILE "IP $IP reached the maximum number of failed attempts!!!\n");
			if(!grep { /$IP/ } @trusted_IPs) {
				print (FILE "Adding IP to the firewall...\n");
				system("$pfctl -t $table -T add $IP/32") == 0 or die "Couldn't add $IP to firewall";
				system("echo $IP/32 >> $tablefile") == 0 or die "Could't write $IP to ${table}'s table file";
			} else {
				print (FILE "...but it is Trusted IP, so i won't block it!\n");
			}
		}
	}
}

close(FILE);
